/* RCS file syntactic analysis */

/******************************************************************************
 *                       Syntax Analysis.
 *                       Keyword table
 *                       Testprogram: define SYNTEST
 *                       Compatibility with Release 2: define COMPAT2=1
 ******************************************************************************
 */

/* Copyright 1982, 1988, 1989 Walter Tichy
   Copyright 1990, 1991, 1992, 1993, 1994, 1995 Paul Eggert
   Distributed under license by the Free Software Foundation, Inc.

This file is part of RCS.

RCS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

RCS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with RCS; see the file COPYING.
If not, write to the Free Software Foundation,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

Report problems and direct all questions to:

    rcs-bugs@cs.purdue.edu

*/

/*
 * $Log: rcssyn.c,v $
 * Revision 6.0  1995/07/28 05:00:00  tompkins
 * updated for rcs 6.0
 *
 * Revision 5.16  1995/07/07  03:09:26  tompkins
 * initial backport of RCSTHREAD code to 5.7.
 *
 * Revision 5.15  1995/06/16 06:19:24  eggert
 * Update FSF address.
 *
 * Revision 5.14  1995/06/01 16:23:43  eggert
 * (expand_names): Add "b" for -kb.
 * (getdelta): Don't strip leading "19" from MKS RCS dates; see cmpdate.
 *
 * Revision 5.13  1994/03/20 04:52:58  eggert
 * Remove lint.
 *
 * Revision 5.12  1993/11/03 17:42:27  eggert
 * Parse MKS RCS dates; ignore \r in diff control lines.
 * Don't discard ignored phrases.  Improve quality of diagnostics.
 *
 * Revision 5.11  1992/07/28  16:12:44  eggert
 * Avoid `unsigned'.  Statement macro names now end in _.
 *
 * Revision 5.10  1992/01/24  18:44:19  eggert
 * Move put routines to rcsgen.c.
 *
 * Revision 5.9  1992/01/06  02:42:34  eggert
 * ULONG_MAX/10 -> ULONG_MAX_OVER_10
 * while (E) ; -> while (E) continue;
 *
 * Revision 5.8  1991/08/19  03:13:55  eggert
 * Tune.
 *
 * Revision 5.7  1991/04/21  11:58:29  eggert
 * Disambiguate names on shortname hosts.
 * Fix errno bug.  Add MS-DOS support.
 *
 * Revision 5.6  1991/02/28  19:18:51  eggert
 * Fix null termination bug in reporting keyword expansion.
 *
 * Revision 5.5  1991/02/25  07:12:44  eggert
 * Check diff output more carefully; avoid overflow.
 *
 * Revision 5.4  1990/11/01  05:28:48  eggert
 * When ignoring unknown phrases, copy them to the output RCS file.
 * Permit arbitrary data in logs and comment leaders.
 * Don't check for nontext on initial checkin.
 *
 * Revision 5.3  1990/09/20  07:58:32  eggert
 * Remove the test for non-text bytes; it caused more pain than it cured.
 *
 * Revision 5.2  1990/09/04  08:02:30  eggert
 * Parse RCS files with no revisions.
 * Don't strip leading white space from diff commands.  Count RCS lines better.
 *
 * Revision 5.1  1990/08/29  07:14:06  eggert
 * Add -kkvl.  Clean old log messages too.
 *
 * Revision 5.0  1990/08/22  08:13:44  eggert
 * Try to parse future RCS formats without barfing.
 * Add -k.  Don't require final newline.
 * Remove compile-time limits; use malloc instead.
 * Don't output branch keyword if there's no default branch,
 * because RCS version 3 doesn't understand it.
 * Tune.  Remove lint.
 * Add support for ISO 8859.  Ansify and Posixate.
 * Check that a newly checked-in file is acceptable as input to 'diff'.
 * Check diff's output.
 *
 * Revision 4.6  89/05/01  15:13:32  narten
 * changed copyright header to reflect current distribution rules
 * 
 * Revision 4.5  88/08/09  19:13:21  eggert
 * Allow cc -R; remove lint.
 * 
 * Revision 4.4  87/12/18  11:46:16  narten
 * more lint cleanups (Guy Harris)
 * 
 * Revision 4.3  87/10/18  10:39:36  narten
 * Updating version numbers. Changes relative to 1.1 actually relative to
 * 4.1
 * 
 * Revision 1.3  87/09/24  14:00:49  narten
 * Sources now pass through lint (if you ignore printf/sprintf/fprintf 
 * warnings)
 * 
 * Revision 1.2  87/03/27  14:22:40  jenkins
 * Port to suns
 * 
 * Revision 4.1  83/03/28  11:38:49  wft
 * Added parsing and printing of default branch.
 * 
 * Revision 3.6  83/01/15  17:46:50  wft
 * Changed readdelta() to initialize selector and log-pointer.
 * Changed puttree to check for selector==DELETE; putdtext() uses DELNUMFORM.
 *
 * Revision 3.5  82/12/08  21:58:58  wft
 * renamed Commentleader to Commleader.
 *
 * Revision 3.4  82/12/04  13:24:40  wft
 * Added routine gettree(), which updates keeplock after reading the
 * delta tree.
 *
 * Revision 3.3  82/11/28  21:30:11  wft
 * Reading and printing of Suffix removed; version COMPAT2 skips the
 * Suffix for files of release 2 format. Fixed problems with printing nil.
 *
 * Revision 3.2  82/10/18  21:18:25  wft
 * renamed putdeltatext to putdtext.
 *
 * Revision 3.1  82/10/11  19:45:11  wft
 * made sure getc() returns into an integer.
 */



/* version COMPAT2 reads files of the format of release 2 and 3, but
 * generates files of release 3 format. Need not be defined if no
 * old RCS files generated with release 2 exist.
 */

#include "rcsbase.h"

libId(synId, "$Id: rcssyn.c,v 6.0 1995/10/17 06:00:00 tompkins Exp $")

static char const *getkeyval P((char const*,enum tokens,int));
static int getdelta P((void));
static int strn2expmode P((char const*,size_t));
static struct hshentry *getdnum P((void));
static void badDiffOutput P((char const*)) exiting;
static void diffLineNumberTooLarge P((char const*)) exiting;
static void getsemi P((char const*));

/* keyword table */

char const
	Kaccess[]   = "access",
	Kauthor[]   = "author",
	Kbranch[]   = "branch",
	Kcomment[]  = "comment",
	Kdate[]     = "date",
	Kdesc[]     = "desc",
	Kexpand[]   = "expand",
	Khead[]     = "head",
	Klocks[]    = "locks",
	Klog[]      = "log",
	Knext[]     = "next",
	Kstate[]    = "state",
	Kstrict[]   = "strict",
	Ksymbols[]  = "symbols",
	Ktext[]     = "text";

static char const
#if COMPAT2
	Ksuffix[]   = "suffix",
#endif
	K_branches[]= "branches";

#ifdef RCSTHREAD
/* define a string with the following RCS file field delimiters:
   BS, HT, NL, VT, FF, CR, SP */
static unsigned char whitespace[] = {
	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x20 };
#endif

static struct buf Commleader;
struct cbuf Comment;
struct cbuf Ignored;
struct access   * AccessList;
struct assoc    * Symbols;
struct rcslock *Locks;
int		  Expand;
int               StrictLocks;
struct hshentry * Head;
char const      * Dbranch;
int TotalDeltas;


	static void
getsemi(key)
	char const *key;
/* Get a semicolon to finish off a phrase started by KEY.  */
{
	if (!getlex(SEMI))
		fatserror("missing ';' after '%s'", key);
}

	static struct hshentry *
getdnum()
/* Get a delta number.  */
{
	register struct hshentry *delta = getnum();
	if (delta && countnumflds(delta->num)&1)
		fatserror("%s isn't a delta number", delta->num);
	return delta;
}


	void
getadmin()
/* Read an <admin> and initialize the appropriate global variables.  */
{
#ifdef REL_TAGS
	char *real_id;
	char *state_ptr;
#endif
	register char const *id;
        struct access   * newaccess;
        struct assoc    * newassoc;
	struct rcslock *newlock;
        struct hshentry * delta;
	struct access **LastAccess;
	struct assoc **LastSymbol;
	struct rcslock **LastLock;
	struct buf b;
	struct cbuf cb;

        TotalDeltas=0;

	getkey(Khead);
	Head = getdnum();
	getsemi(Khead);

	Dbranch = 0;
	if (getkeyopt(Kbranch)) {
		if ((delta = getnum()))
			Dbranch = delta->num;
		getsemi(Kbranch);
        }


#if COMPAT2
        /* read suffix. Only in release 2 format */
	if (getkeyopt(Ksuffix)) {
                if (nexttok==STRING) {
			readstring(); nextlex(); /* Throw away the suffix.  */
		} else if (nexttok==ID) {
                        nextlex();
                }
		getsemi(Ksuffix);
        }
#endif

	getkey(Kaccess);
	LastAccess = &AccessList;
	while ((id = getid())) {
		newaccess = ftalloc(struct access);
                newaccess->login = id;
		*LastAccess = newaccess;
		LastAccess = &newaccess->nextaccess;
        }
	*LastAccess = 0;
	getsemi(Kaccess);

	getkey(Ksymbols);
	LastSymbol = &Symbols;
        while ((id = getid())) {
                if (!getlex(COLON))
			fatserror("missing ':' in symbolic name definition");
                if (!(delta=getnum())) {
			fatserror("missing number in symbolic name definition");
                } else { /*add new pair to association list*/
			newassoc = ftalloc(struct assoc);
#ifdef REL_TAGS
/* "id" is a temporary copy of the scanned in symbol.
   Intercept the parser here to check for released keyword */
			if ((real_id=has_tag_state(id,&state_ptr)) != NULL) {
				newassoc->symbol=real_id; /* update released tag list */
				newassoc->tag_state=state_ptr;
			} else {
				newassoc->symbol=id;
				newassoc->tag_state=NULL;
			}
#else
                        newassoc->symbol=id;
#endif /* REL_TAGS */
			newassoc->num = delta->num;
			*LastSymbol = newassoc;
			LastSymbol = &newassoc->nextassoc;
                }
        }
	*LastSymbol = 0;
	getsemi(Ksymbols);

	getkey(Klocks);
	LastLock = &Locks;
        while ((id = getid())) {
                if (!getlex(COLON))
			fatserror("missing ':' in lock");
		if (!(delta=getdnum())) {
			fatserror("missing number in lock");
                } else { /*add new pair to lock list*/
			newlock = ftalloc(struct rcslock);
                        newlock->login=id;
                        newlock->delta=delta;
			*LastLock = newlock;
			LastLock = &newlock->nextlock;
                }
        }
	*LastLock = 0;
	getsemi(Klocks);

	if ((StrictLocks = getkeyopt(Kstrict)))
		getsemi(Kstrict);

	clear_buf(&Comment);
	if (getkeyopt(Kcomment)) {
		if (nexttok==STRING) {
			Comment = savestring(&Commleader);
			nextlex();
		}
		getsemi(Kcomment);
        }

	Expand = KEYVAL_EXPAND;
	if (getkeyopt(Kexpand)) {
		if (nexttok==STRING) {
			bufautobegin(&b);
			cb = savestring(&b);
			if ((Expand = strn2expmode(cb.string,cb.size)) < 0)
			    fatserror("unknown expand mode %.*s",
				(int)cb.size, cb.string
			    );
			bufautoend(&b);
			nextlex();
		}
		getsemi(Kexpand);
        }
	Ignored = getphrases(Kdesc);
}

#ifdef REL_TAGS
char *has_tag_state(symbol_name, state_ptr)
/* char const *symbol_name; */
char *symbol_name;
char **state_ptr;
{
	char *real_symbol, *end_token_ptr, *tmp_state_ptr;
	int real_symbol_len, real_state_len;

	*state_ptr=NULL;
	if (strlen(symbol_name) < (2*strlen(TAG_STATE_TOKEN)+1))
		return(NULL);
	if((tmp_state_ptr=strstr(symbol_name,TAG_STATE_TOKEN))==NULL)
		return(NULL);
	real_symbol_len=strlen(symbol_name)-strlen(tmp_state_ptr);
	end_token_ptr=symbol_name+strlen(symbol_name)-strlen(TAG_STATE_TOKEN);
	if (strcmp(TAG_STATE_TOKEN,end_token_ptr)==0) {
		/* make sure there are two token strings and that they
		   don't overlap */
		if (tmp_state_ptr+strlen(TAG_STATE_TOKEN) >= end_token_ptr)
			return(NULL);
		/* adjust tmp_state_ptr to 1st char beyond beginning delimiter */
		(tmp_state_ptr)+=strlen(TAG_STATE_TOKEN);
		/* copy state name */
		real_state_len=strlen(tmp_state_ptr)-strlen(TAG_STATE_TOKEN);
		*state_ptr=ftnalloc(char,strlen(tmp_state_ptr));
		strncpy(*state_ptr,tmp_state_ptr,real_state_len);
		*((*state_ptr)+real_state_len) = '\0';
		/* copy actual symbol name */
		real_symbol=ftnalloc(char,strlen(symbol_name));
		strncpy(real_symbol,symbol_name,real_symbol_len);
		real_symbol[real_symbol_len]='\0';
		return(real_symbol);
	} else
		return(NULL);
}
#endif REL_TAGS

char const *const expand_names[] = {
	/* These must agree with *_EXPAND in rcsbase.h.  */
	"kv", "kvl", "k", "v", "o", "b",
	0
};

	int
str2expmode(s)
	char const *s;
/* Yield expand mode corresponding to S, or -1 if bad.  */
{
	return strn2expmode(s, strlen(s));
}

	static int
strn2expmode(s, n)
	char const *s;
	size_t n;
{
	char const *const *p;

	for (p = expand_names;  *p;  ++p)
		if (memcmp(*p,s,n) == 0  &&  !(*p)[n])
			return p - expand_names;
	return -1;
}


	void
ignorephrases(key)
	const char *key;
/*
* Ignore a series of phrases that do not start with KEY.
* Stop when the next phrase starts with a token that is not an identifier,
* or is KEY.
*/
{
	for (;;) {
		nextlex();
		if (nexttok != ID  ||  strcmp(NextString,key) == 0)
			break;
		warnignore();
		hshenter=false;
		for (;; nextlex()) {
			switch (nexttok) {
				case SEMI: hshenter=true; break;
				case ID:
				case NUM: ffree1(NextString); continue;
				case STRING: readstring(); continue;
				default: continue;
			}
			break;
		}
	}
}


	static int
getdelta()
/* Function: reads a delta block.
 * returns false if the current block does not start with a number.
 */
{
        register struct hshentry * Delta, * num;
	struct branchhead **LastBranch, *NewBranch;

	if (!(Delta = getdnum()))
		return false;

        hshenter = false; /*Don't enter dates into hashtable*/
	Delta->date = getkeyval(Kdate, NUM, false);
        hshenter=true;    /*reset hshenter for revision numbers.*/

        Delta->author = getkeyval(Kauthor, ID, false);

        Delta->state = getkeyval(Kstate, ID, true);

	getkey(K_branches);
	LastBranch = &Delta->branches;
	while ((num = getdnum())) {
		NewBranch = ftalloc(struct branchhead);
                NewBranch->hsh = num;
		*LastBranch = NewBranch;
		LastBranch = &NewBranch->nextbranch;
        }
	*LastBranch = 0;
	getsemi(K_branches);

	getkey(Knext);
	Delta->next = num = getdnum();
	getsemi(Knext);
	Delta->lockedby = 0;
	Delta->log.string = 0;
	Delta->selector = true;
	Delta->ig = getphrases(Kdesc);
        TotalDeltas++;
        return (true);
}


	void
gettree()
/* Function: Reads in the delta tree with getdelta(), then
 * updates the lockedby fields.
 */
{
	struct rcslock const *currlock;

	while (getdelta())
		continue;
        currlock=Locks;
        while (currlock) {
                currlock->delta->lockedby = currlock->login;
                currlock = currlock->nextlock;
        }
}


	void
getdesc(prdesc)
int  prdesc;
/* Function: read in descriptive text
 * nexttok is not advanced afterwards.
 * If prdesc is set, the text is printed to stdout.
 */
{

	getkeystring(Kdesc);
        if (prdesc)
                printstring();  /*echo string*/
        else    readstring();   /*skip string*/
}






	static char const *
getkeyval(keyword, token, optional)
	char const *keyword;
	enum tokens token;
	int optional;
/* reads a pair of the form
 * <keyword> <token> ;
 * where token is one of <id> or <num>. optional indicates whether
 * <token> is optional. A pointer to
 * the actual character string of <id> or <num> is returned.
 */
{
	register char const *val = 0;

	getkey(keyword);
        if (nexttok==token) {
                val = NextString;
                nextlex();
        } else {
		if (!optional)
			fatserror("missing %s", keyword);
        }
	getsemi(keyword);
        return(val);
}

#ifdef RCSTHREAD
char *eval_thread(threadfile,rcsfilename,statusmsg,stflagptr)
char *threadfile;
char *rcsfilename;
char *statusmsg;
int  *stflagptr;
{
	int found, search_status;
	int use_latest=false;
	struct thr_struct *curr_thr_symlist, *thread_symlist;
	char *RCSbasename, *thread_line=NULL;
	char *matchname, *current_sfx;

	RCSbasename=basename(rcsfilename);
	if((current_sfx=rcssuffix(RCSname)) == (char *) 0)
		current_sfx=",v";
	if((thread_symlist=getthr_symlist(threadfile,RCSbasename,current_sfx,&thread_line,statusmsg,stflagptr))==NULL)
		return(NULL);
	found=false;
	curr_thr_symlist=thread_symlist;
	while (curr_thr_symlist!=NULL) {
		if (strcmp(curr_thr_symlist->symbol,LATEST_REV)==0) {
			found=true;
			use_latest=true;
			break;
		}
		if (is_revnum(curr_thr_symlist->symbol)) {
			if (scan_rev_tree(Head,curr_thr_symlist->symbol)) {
				found=true;
				break;	/* found a match in the revision tree */
			}
		}
		if (scan_symbol_list(Symbols,curr_thr_symlist->symbol)) {
			found=true;
			break;
		}
		curr_thr_symlist=curr_thr_symlist->next_thr;
	}
	if(found) {
		if (use_latest) {
			sprintf(statusmsg, "The thread file \"%s\" indicates that the latest revision\non the default branch should be selected for the file \"%s\".\n", threadfile, RCSname);
		}
		*stflagptr=THR_FOUND;
		if((matchname=(char *) malloc(strlen(curr_thr_symlist->symbol)+1))==NULL){
			sprintf(statusmsg
				,"Memory allocation error parsing thread in line %d of %s\n"
				,__LINE__,__FILE__);
			goto cleanup;
		}
		strcpy(matchname,curr_thr_symlist->symbol);
		if (use_latest)
			sprintf(statusmsg, "The thread file \"%s\" indicates that the latest revision\non the default branch should be selected for the file \"%s\".\n", threadfile, RCSname);
		else
			sprintf(statusmsg,"The thread file \"%s\" indicates that symbol/revision \"%s\"\nshould be selected for the file \"%s\".\n", threadfile, curr_thr_symlist->symbol, RCSname);
	} else {
		matchname=NULL;
		*stflagptr=THR_NOMATCH;
		sprintf(statusmsg,"Thread file \"%s\" contains no matching symbol/revision\nfor the file \"%s\".\n",threadfile,RCSname);
	}
	/* all done with thread list, free up the space */
cleanup:
	free_thr_symlist(thread_symlist);
	if (thread_line != NULL) free(thread_line);
	return(matchname);
}

int scan_rev_tree(root,thr_rev_entry)
struct hshentry const *root;
char  *thr_rev_entry;
/* Function: Searches revision tree for a revision matching "thr_rev_entry".
 */
{
	struct branchhead const *nextbranch;
	int thr_entry_len;

	if (root==NULL) return(false);

	if (root->selector) {
		/* Is the passed-in thread revision entry the current item in
		   the revision tree? */
		thr_entry_len=strlen(thr_rev_entry);
		if (strncmp(thr_rev_entry,root->num,thr_entry_len)==0) {
			/* we know we have either an exact match or a subset */
			if(thr_entry_len != strlen(root->num)) { /* ie. 1.1. vs. 1.1.1.1 */
				if(thr_rev_entry[thr_entry_len-1] == '.')
					return(true); /* branch head match OK, ie. thr_entry='1.1.' */
			} else {
				return(true); /* Exact Match between thr_entry and rev. */
			}
		}
	}

	if(scan_rev_tree(root->next,thr_rev_entry))
		return(true);

	nextbranch = root->branches;
	while (nextbranch) {
		if(scan_rev_tree(nextbranch->hsh,thr_rev_entry))
			return(true);
		nextbranch = nextbranch->nextbranch;
	}
	return(false);
}

int scan_symbol_list(symlist,thr_entry)
struct assoc *symlist;
char *thr_entry;
{
	while (symlist!=NULL) {
		if (strcmp(thr_entry,symlist->symbol)==0)
			return(true);
		symlist=symlist->nextassoc;
	}
	return(false);
}

int is_revnum(thr_entry)
char *thr_entry;
{
	int pos;

	pos=0;
	while (thr_entry[pos] != '\0') {
		if (!isdigit(thr_entry[pos]) && thr_entry[pos]!='.')
			return(false);
		++pos;
	}
	return(true);
}

	struct thr_struct
*getthr_symlist(threadfile,rcsbasefile,current_sfx,trg_thr_line
                ,statusmsg,stflagptr)
char *threadfile;
char *rcsbasefile;
char *current_sfx;
char **trg_thr_line;
char *statusmsg;
int  *stflagptr;
{
	struct thr_struct *thr_symlist;
	FILE *thread_fp;
	int fn_len, rcs_name_len, sfx_len, found;
	char file_name[SIZEABLE_PATH];
	char *default_line, *scratchname, *thread_line=NULL, *colon_pos, *firstwhitespc=NULL;
	char *tmp_rcs, *tmp_fn, *filename_sfx_ptr, *rcsbfile_sfx_ptr=NULL;
	
	rcs_name_len=strlen(rcsbasefile);
	sfx_len=strlen(current_sfx);
	found=false;
	thr_symlist=NULL;
	default_line=NULL;
	*stflagptr=THR_FOUND; /* This must be init'ed so cleanup: can check it */
	if ((thread_fp=fopen(threadfile,"r")) == NULL) {
		sprintf(statusmsg,"Unable to access thread file \"%s\".\n",threadfile);
		*stflagptr=THR_NOFILE;
		goto cleanup;
	}
	/* call semi-dynamic version of fgets with a reasonable guess of how
		long a line in the thread file might be.  Unless an alloc error occurs,
		function guarantees that entire lines in the thread file will be
		read in. */
	while ((thread_line=dyn_fgets(SIZEABLE_PATH+100,50,thread_fp)) != NULL) {
		if(strlen(thread_line) < 1) {
			/* Ignore blank lines in thread file */
			free(thread_line);
			continue;
		}
		if ((colon_pos=strchr(thread_line,':')) == NULL) {
			sprintf(statusmsg
				,"Thread parse error - no ':' delimiter in thread entry:\n\"%s\"\n"
				,thread_line);
			*stflagptr=THR_PARSERR;
			goto cleanup;
		}
/* Lose extra whitespace before colon. */
		if ((firstwhitespc=strpbrk(thread_line,whitespace))==NULL
		 || firstwhitespc>colon_pos)
			fn_len=colon_pos-thread_line;
		else
			fn_len=firstwhitespc-thread_line;
		strncpy(file_name,thread_line,fn_len); file_name[fn_len]='\0';
		filename_sfx_ptr=file_name+(fn_len-(sfx_len>fn_len?fn_len:sfx_len));
		if (strcmp(filename_sfx_ptr,current_sfx) != 0)
			filename_sfx_ptr=file_name+fn_len;
		rcsbfile_sfx_ptr=rcsbasefile+rcs_name_len
		  -(sfx_len>rcs_name_len?rcs_name_len:sfx_len);
		if (strcmp(rcsbfile_sfx_ptr,current_sfx) != 0)
			rcsbfile_sfx_ptr=rcsbasefile+rcs_name_len;
		if((tmp_fn=malloc(fn_len+1))==NULL) faterror("memory alloc failed");
		if((tmp_rcs=malloc(rcs_name_len+1))==NULL)faterror("memory alloc failed");
		strncpy(tmp_fn,file_name,filename_sfx_ptr-file_name);
		tmp_fn[filename_sfx_ptr-file_name]='\0';
		strncpy(tmp_rcs,rcsbasefile,rcsbfile_sfx_ptr-rcsbasefile);
		tmp_rcs[rcsbfile_sfx_ptr-rcsbasefile]='\0';
		if (strcmp(tmp_fn,tmp_rcs)==0) {
			free(tmp_fn); free(tmp_rcs);
#if defined(RCSTHREAD) && defined(NODUPE_THR)
				if (found) {
					sprintf(statusmsg
						,"Thread error - duplicate entry for %s encountered.\n"
						,file_name);
					*stflagptr=THR_PARSERR;
					goto cleanup;
				} else {
					found=true;
					*trg_thr_line=thread_line;
					continue;
				}
#else
				found=true;
				*trg_thr_line=thread_line;
				break;	/* found the correct entry in the thread file */
#endif
		} else {
			free(tmp_fn); free(tmp_rcs);
		}

		if (strncmp(file_name,DEF_THR_ENTRY,strlen(DEF_THR_ENTRY))==0){
			/* See if current entry matches "DEFAULT" */
			scratchname=ftnalloc(char,
				strlen(DEF_THR_ENTRY)+strlen(current_sfx)+1);
			strcpy(scratchname,DEF_THR_ENTRY);
			if (strcmp(scratchname,file_name)==0) {
				default_line=thread_line;
				continue;	/* prevent free() from wiping out this "thread_line" */
			} else {
		/* See if current entry matches "DEFAULT,v", or whatever the suffix is. */
				strcat(scratchname,current_sfx);
				if (strcmp(scratchname,file_name)==0) {
					default_line=thread_line;
					continue; /* prevent free() from wiping out this "thread_line" */
				}
			}
		}
free_thr_entry:
		free(thread_line);
	}
	/* If "found" is true, the requested thread item is in "thread_line" */
	if (!found)	{ /* We'll have to use the default entry */
		if (default_line != NULL) {
			*trg_thr_line=default_line; /* save this so it can be freed by caller */
			sprintf(statusmsg
				,"Requested RCS file \"%s\" not in thread file \"%s\".\nUsing DEFAULT thread entry.\n",rcsbasefile,threadfile);
			if((thr_symlist=parse_thread_entry(default_line)) == NULL) {
				strcat(statusmsg,"Error parsing DEFAULT thread entry,\n");
				*stflagptr=THR_PARSERR;
				goto cleanup;
			} 
		} else {	/* No matching entry found and No DEFAULT entry */
			*trg_thr_line=NULL; /* no entry to be freed by caller */
			sprintf(statusmsg
				,"Unable to find thread entry for \"%s\" in %s.\n"
				,rcsbasefile,threadfile);
				*stflagptr=THR_NOENTRY;
			goto cleanup;
		}
	} else {/* We have a match, the default entry (if any) is no longer needed */
		if(default_line != NULL)
			free(default_line);
		if((thr_symlist=parse_thread_entry(*trg_thr_line)) == NULL) {
			sprintf(statusmsg,"Error parsing thread entry,\n\"%s\",\nin \"%s\"\n"
				,trg_thr_line,threadfile);
			*stflagptr=THR_PARSERR;
			goto cleanup;
		}
	}
cleanup:
	if(thread_fp!=NULL)
		fclose(thread_fp);
	if(*stflagptr != THR_FOUND) {
		if(default_line != NULL)
			free(default_line);
		if(thread_line != NULL)
			free(thread_line);
	}
	return(thr_symlist);
}

struct thr_struct *parse_thread_entry(thread_line)
char *thread_line;
{
	struct thr_struct *thr_symlist, *head_thr_symlist, *next_thr_symlist;
	char *colon_ptr, *tokenptr;

/* Find ":" which separates RCS filename entry from tags
	Create head of thread symbol list
	Save the pointer to the head of the list
	Find the first token (tag) in the thread entry
	Store this entry in the list
	Create the next link in the list
	Move current link pointer ahead
	Find and store the rest of the tags in the list */

	if((colon_ptr=strchr(thread_line,':')) == NULL)
		return(NULL);
	if((tokenptr=strtok(colon_ptr+1,whitespace)) == NULL)
		return(NULL);
	if((thr_symlist=(struct thr_struct *) malloc(sizeof(struct thr_struct))) == NULL)
		return(NULL);
	head_thr_symlist=thr_symlist;
	head_thr_symlist->prev_thr=NULL;
	head_thr_symlist->next_thr=NULL;
	thr_symlist->symbol=tokenptr;

	while ((tokenptr=strtok(NULL,whitespace)) != NULL) {
		if((next_thr_symlist=(struct thr_struct *) malloc(sizeof(struct thr_struct)))==NULL){
			free_thr_symlist(head_thr_symlist);
			return(NULL);
		}
		thr_symlist->next_thr=next_thr_symlist;
		next_thr_symlist->prev_thr=thr_symlist;
		thr_symlist=next_thr_symlist;
		thr_symlist->symbol=tokenptr;
		thr_symlist->next_thr=NULL;
	}
	return(head_thr_symlist);
}

void free_thr_symlist(head_thr_symlist)
struct thr_struct *head_thr_symlist;
{
	struct thr_struct *curr_thr_symlist;

	/* First go to the end of the linked list of records, so that we
		can destroy it from the end to the beginning. */
	curr_thr_symlist=head_thr_symlist;
	while (curr_thr_symlist->next_thr != NULL)
		curr_thr_symlist=curr_thr_symlist->next_thr;
	/* Now that we are at the end of the list, backup and free the next record
		until we reach the beginning of the list */
	while (curr_thr_symlist->prev_thr != NULL) {
		curr_thr_symlist=curr_thr_symlist->prev_thr;
		free(curr_thr_symlist->next_thr);
	}
	free(head_thr_symlist);
	return;
}
#endif

	void
unexpected_EOF()
{
	rcsfaterror("unexpected EOF in diff output");
}

	void
initdiffcmd(dc)
	register struct diffcmd *dc;
/* Initialize *dc suitably for getdiffcmd(). */
{
	dc->adprev = 0;
	dc->dafter = 0;
}

	static void
badDiffOutput(buf)
	char const *buf;
{
	rcsfaterror("bad diff output line: %s", buf);
}

	static void
diffLineNumberTooLarge(buf)
	char const *buf;
{
	rcsfaterror("diff line number too large: %s", buf);
}

	int
getdiffcmd(finfile, delimiter, foutfile, dc)
	RILE *finfile;
	FILE *foutfile;
	int delimiter;
	struct diffcmd *dc;
/* Get a editing command output by 'diff -n' from fin.
 * The input is delimited by SDELIM if delimiter is set, EOF otherwise.
 * Copy a clean version of the command to fout (if nonnull).
 * Yield 0 for 'd', 1 for 'a', and -1 for EOF.
 * Store the command's line number and length into dc->line1 and dc->nlines.
 * Keep dc->adprev and dc->dafter up to date.
 */
{
	register int c;
	declarecache;
	register FILE *fout;
	register char *p;
	register RILE *fin;
	long line1, nlines, t;
	char buf[BUFSIZ];

	fin = finfile;
	fout = foutfile;
	setupcache(fin); cache(fin);
	cachegeteof_(c, { if (delimiter) unexpected_EOF(); return -1; } )
	if (delimiter) {
		if (c==SDELIM) {
			cacheget_(c)
			if (c==SDELIM) {
				buf[0] = c;
				buf[1] = 0;
				badDiffOutput(buf);
			}
			uncache(fin);
			nextc = c;
			if (fout)
				aprintf(fout, "%c%c", SDELIM, c);
			return -1;
		}
	}
	p = buf;
	do {
		if (buf+BUFSIZ-2 <= p) {
			rcsfaterror("diff output command line too long");
		}
		*p++ = c;
		cachegeteof_(c, unexpected_EOF();)
	} while (c != '\n');
	uncache(fin);
	if (delimiter)
		++rcsline;
	*p = '\0';
	for (p = buf+1;  (c = *p++) == ' ';  )
		continue;
	line1 = 0;
	while (isdigit(c)) {
		if (
			LONG_MAX/10 < line1  ||
			(t = line1 * 10,   (line1 = t + (c - '0'))  <  t)
		)
			diffLineNumberTooLarge(buf);
		c = *p++;
	}
	while (c == ' ')
		c = *p++;
	nlines = 0;
	while (isdigit(c)) {
		if (
			LONG_MAX/10 < nlines  ||
			(t = nlines * 10,   (nlines = t + (c - '0'))  <  t)
		)
			diffLineNumberTooLarge(buf);
		c = *p++;
	}
	if (c == '\r')
		c = *p++;
	if (c || !nlines) {
		badDiffOutput(buf);
	}
	if (line1+nlines < line1)
		diffLineNumberTooLarge(buf);
	switch (buf[0]) {
	    case 'a':
		if (line1 < dc->adprev) {
		    rcsfaterror("backward insertion in diff output: %s", buf);
		}
		dc->adprev = line1 + 1;
		break;
	    case 'd':
		if (line1 < dc->adprev  ||  line1 < dc->dafter) {
		    rcsfaterror("backward deletion in diff output: %s", buf);
		}
		dc->adprev = line1;
		dc->dafter = line1 + nlines;
		break;
	    default:
		badDiffOutput(buf);
	}
	if (fout) {
		aprintf(fout, "%s\n", buf);
	}
	dc->line1 = line1;
	dc->nlines = nlines;
	return buf[0] == 'a';
}



#ifdef SYNTEST

/* Input an RCS file and print its internal data structures.  */

char const cmdid[] = "syntest";

	int
main(argc,argv)
int argc; char * argv[];
{

        if (argc<2) {
		aputs("No input file\n",stderr);
		exitmain(EXIT_FAILURE);
        }
	if (!(finptr = Iopen(argv[1], FOPEN_R, (struct stat*)0))) {
		faterror("can't open input file %s", argv[1]);
        }
        Lexinit();
        getadmin();
	fdlock = STDOUT_FILENO;
	putadmin();

        gettree();

        getdesc(true);

	nextlex();

	if (!eoflex()) {
		fatserror("expecting EOF");
        }
	exitmain(EXIT_SUCCESS);
}

void exiterr() { _exit(EXIT_FAILURE); }

#endif
